/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008,2012  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "ConfigDialog.hpp"
#include "GameWindow.hpp"

#include "Fractal.hpp"
#include "CSet.hpp"
#include "Checker.hpp"

#include <cstdlib>

namespace islandmandel
{
namespace gimandel
{

/* ************************************************************************** */
/* **************************** ConfigDialog ******************************** */
/* ************************************************************************** */

// Fractal-sizes for game and preview
#define GAME_WIDTH 640
#define GAME_HEIGHT 480
#define PREVIEW_WIDTH 120
#define PREVIEW_HEIGHT 92
#define JULIAC_WIDTH 120
#define JULIAC_HEIGHT 92

ConfigDialog::ConfigDialog(GameWindow& g)
  : game(g),
    dlg(gtk_dialog_new_with_buttons("Choose Fractal",
      GTK_WINDOW(game.getWidget()), GTK_DIALOG_MODAL,
      GTK_STOCK_OK, GTK_RESPONSE_OK,
      GTK_STOCK_CANCEL, GTK_RESPONSE_NONE,
      NULL)),
    radioChecker(gtk_radio_button_new_with_label(NULL, "Checker-Board")),
    radioMandelbrot(gtk_radio_button_new_with_label_from_widget(
      GTK_RADIO_BUTTON(radioChecker),
      "Mandelbrot")),
    radioJulia(gtk_radio_button_new_with_label_from_widget(
      GTK_RADIO_BUTTON(radioChecker),
      "Julia")),
    juliaBox(gtk_vbox_new(FALSE, 2)),
    juliaCX(gtk_entry_new()), juliaCY(gtk_entry_new()),
    juliaCMandelFract(new Mandelbrot(JULIAC_WIDTH, JULIAC_HEIGHT)),
    juliaCMandel(),
    previewFract(NULL), preview()
{
  assert(dlg);

  // Set up the dialog window
  gtk_window_set_resizable(GTK_WINDOW(dlg), FALSE);

  // Create the basic vertical split into preview and config area
  GtkWidget* hbox(gtk_hbox_new(TRUE, 5));
  assert(hbox);
  gtk_box_pack_end(GTK_BOX(hbox), preview.getWidget(), FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), hbox, TRUE, TRUE, 0);

  // Initial configuration is Mandelbrot, before signal connection!
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radioMandelbrot), TRUE);
  updatePreview(false);

  // Build up the Julia-configuration part
  assert(juliaBox && juliaCX && juliaCY);
  GtkWidget* juliaEntries(gtk_hbox_new(FALSE, 2));
  #define CONFIG_JULIA_ENTRY(name) \
    gtk_entry_set_width_chars(GTK_ENTRY(name), 4); \
    g_signal_connect(G_OBJECT(name), "changed", \
      G_CALLBACK(&handle_juliaEntry), this);
  CONFIG_JULIA_ENTRY(juliaCX)
  CONFIG_JULIA_ENTRY(juliaCY)
  #define ADD_JULIA_ENTRY(ent) \
    gtk_box_pack_start(GTK_BOX(juliaEntries), ent, FALSE, FALSE, 0);
  ADD_JULIA_ENTRY(gtk_label_new("c="))
  ADD_JULIA_ENTRY(juliaCX)
  ADD_JULIA_ENTRY(gtk_label_new("+"))
  ADD_JULIA_ENTRY(juliaCY)
  ADD_JULIA_ENTRY(gtk_label_new("i"))
  gtk_box_pack_start(GTK_BOX(juliaBox), juliaEntries, FALSE, FALSE, 0);
  assert(juliaCMandelFract);
  juliaCMandelFract->calculate();
  juliaCMandel.setFractal(*juliaCMandelFract);
  gtk_widget_add_events(juliaCMandel.getWidget(), GDK_BUTTON_PRESS_MASK);
  g_signal_connect(G_OBJECT(juliaCMandel.getWidget()), "button_press_event",
    G_CALLBACK(&handle_juliaCClick), this);
  gtk_box_pack_start(GTK_BOX(juliaBox), juliaCMandel.getWidget(),
    FALSE, FALSE, 0);

  // Put the radio-buttons into a box, connect the toggled-signal and add them
  GtkWidget* radioBox(gtk_vbox_new(FALSE, 5));
  assert(radioChecker && radioMandelbrot && radioJulia);
  #define ADD_RADIO(name) \
    gtk_box_pack_start(GTK_BOX(radioBox), radio##name, FALSE, FALSE, 0); \
    g_signal_connect(G_OBJECT(radio##name), "toggled", \
      G_CALLBACK(&handle_radioToggle), this);
  ADD_RADIO(Checker)
  ADD_RADIO(Mandelbrot)
  ADD_RADIO(Julia)
  gtk_box_pack_start(GTK_BOX(radioBox), juliaBox, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), radioBox, TRUE, TRUE, 0);

  // Show them all
  gtk_widget_show_all(hbox);

  // Hide Julia-configuration box
  gtk_widget_hide(juliaBox);
}

ConfigDialog::~ConfigDialog()
{
  assert(juliaCMandelFract);
  delete juliaCMandelFract;

  if(previewFract)
    delete previewFract;
}

bool ConfigDialog::run()
{
  assert(dlg);
  const gint response(gtk_dialog_run(GTK_DIALOG(dlg)));
  gtk_widget_hide(dlg);

  // If ok, configure game with data from here
  if(response==GTK_RESPONSE_OK)
  {
    game.setFractal(buildChosenFractal(GAME_WIDTH, GAME_HEIGHT));
    return true;
  }

  // No, cancelled
  return false;
}

void ConfigDialog::updatePreview(bool redraw)
{
  if(previewFract)
    delete previewFract;
  previewFract=buildChosenFractal(PREVIEW_WIDTH, PREVIEW_HEIGHT);
  assert(previewFract);
  previewFract->calculate();
  preview.setFractal(*previewFract);

  if(redraw)
    preview.redraw();
}

Fractal* ConfigDialog::buildChosenFractal(
  Coordinate::compT w, Coordinate::compT h) const
{
  // Checker?
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(radioChecker)))
    return new Checker(w, h);

  // Mandelbrot?
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(radioMandelbrot)))
    return new Mandelbrot(w, h);

  // Julia?
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(radioJulia)))
  {
    assert(juliaCX && juliaCY);
    return new Julia(w, h,
      std::atof(gtk_entry_get_text(GTK_ENTRY(juliaCX))),
      std::atof(gtk_entry_get_text(GTK_ENTRY(juliaCY))));
  }

  // Nothing right!
  assert(false);
}

#define REVIVE_ME \
  ConfigDialog& me(*static_cast<ConfigDialog*>(ptr));

void ConfigDialog::handle_radioToggle(GtkWidget* w, gpointer ptr)
{
  REVIVE_ME
  assert(w==me.radioChecker || w==me.radioMandelbrot || w==me.radioJulia);

  const bool isActive(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w)));

  // When Julia-button was toggled, show/hide Julia box
  if(w==me.radioJulia)
  {
    assert(me.juliaBox);
    if(isActive)
      gtk_widget_show(me.juliaBox);
    else
      gtk_widget_hide(me.juliaBox);
  }
  
  // Only update once, that is, if "we" were toggled
  if(isActive)
    me.updatePreview(true);
}

void ConfigDialog::handle_juliaEntry(GtkWidget* w, gpointer ptr)
{
  REVIVE_ME
  assert(w==me.juliaCX || w==me.juliaCY);

  me.updatePreview(true);
}

gboolean ConfigDialog::handle_juliaCClick(GtkWidget* w, GdkEventButton* evt,
  gpointer ptr)
{
  REVIVE_ME
  assert(w==me.juliaCMandel.getWidget());

  // Get the coordinates of the click
  const Coordinate clickPos(
    Coordinate::round(evt->x),
    Coordinate::round(evt->y));
  assert(me.juliaCMandelFract);
  const float rx(me.juliaCMandelFract->getRealX(clickPos.getX()));
  const float ry(me.juliaCMandelFract->getRealY(clickPos.getY()));

  // Convert to text and set the entries
  char str[6]; // sign, 1 digit, comma, 2 digits, 0 = 6 characters
  snprintf(str, sizeof(str), "%1.2f", rx);
  gtk_entry_set_text(GTK_ENTRY(me.juliaCX), str);
  snprintf(str, sizeof(str), "%1.2f", ry);
  gtk_entry_set_text(GTK_ENTRY(me.juliaCY), str);

  // Update of the perview is done automatically by the emitted changed-signal

  return TRUE;
}

} // namespace gimandel
} // namespace islandmandel
