/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef GIMANDEL_FRACTALDISPLAY_HPP
#define GIMANDEL_FRACTALDISPLAY_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Fractal.hpp"

#include <gtk/gtk.h>

#include <cassert>

namespace islandmandel
{
namespace gimandel
{

/* ************************************************************************** */
/* ************************** FractalDisplay ******************************** */
/* ************************************************************************** */

/**
 * A Gtk+ widget wrapper capable of displaying fractals; also the main map
 * widget is a subclass, that additionally adds drawing of the figure.
 */
class FractalDisplay
{

  /* ************************************************************************ */
  public:

    /**
     * Create it without fractal at the moment
     */
    FractalDisplay();

    /**
     * Destroy, freeing the colour map
     */
    virtual ~FractalDisplay();

    /**
     * Set/replace the fractal, must be already calculated!
     * @param fr The new Fractal
     */
    void setFractal(const Fractal& fr);

    /**
     * Query for the widget
     * @return The widget representing our map
     */
    inline GtkWidget* getWidget()
    {
      assert(canv);
      return canv;
    }

    /**
     * Request a re-draw
     */
    void redraw() const;

  /* ************************************************************************ */
  protected:

    /**
     * Access the fractal
     * @return The fractal
     */
    inline const Fractal& getFractal() const
    {
      assert(fract);
      return *fract;
    }

    /**
     * Map a colour-code to RGB
     * @param col The colour-code
     * @return The colour in RGB
     */
    inline guint32 getRGB(Figure::colourT col) const
    {
      assert(colourMap);
      assert(fract);
      assert(col<fract->getNumberColours());
      return colourMap->colors[col];
    }

    /**
     * Here the actual drawing on a Drawable is done; this can be overridden
     * to change/extend drawing behaviour.
     * @param screen The drawable to draw to
     * @param gc The graphics context
     */
    virtual void draw(GdkDrawable* screen, GdkGC* gc) const;

  /* ************************************************************************ */
  private:

    /** The Gtk+ widget we use as frontend */
    GtkWidget* canv;

    /** Reference to our Fractal */
    const Fractal* fract;

    /** The colour-map mapping the colour-codes to RGB */
    GdkRgbCmap* colourMap;

    /**
     * Do the drawing when exposed
     */
    void draw() const;

    /**
     * Handler routine for the exposed event, calls into redraw.
     * @param w The widget, should be ptr->canv
     * @param evt The expose event data
     * @param ptr this-pointer
     */
    static gboolean on_expose(GtkWidget* w, GdkEventExpose* evt, gpointer ptr);

};

} // namespace gimandel
} // namespace islandmandel

#endif // Header-guard
