/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "MapDisplay.hpp"

#include "Figure.hpp"
#include "Fractal.hpp"

namespace islandmandel
{
namespace gimandel
{

/* ************************************************************************** */
/* **************************** MapDisplay ********************************** */
/* ************************************************************************** */

const Coordinate::compT MapDisplay::FIGURE_DOT_RADIUS(8);

void MapDisplay::draw(GdkDrawable* screen, GdkGC* gc) const
{
  // Draw the fractal
  FractalDisplay::draw(screen, gc);

  // Plot the figure-dots
  assert(fig); // Should have been set prior to drawing!
  for(Figure::Iterator i(*fig); i; ++i)
  {
    /**
     * A figure-dot is displayed as a circle-line of its own colour and a
     * connection-line to the Figure's center in red/green depending on if the
     * dot matches or not.  The dots circumference is also drawn "below" its
     * colour in red/green.
     */

    const Coordinate& coord(i.getCoord());

    /* Do the red/green stuff */

    // Find correct colour
    GdkColor rgCol;
    if(getFractal().getPixelColour(coord)==i.getColour())
    {
      rgCol.red=0x0000;
      rgCol.green=0xFFFF;
      rgCol.blue=0x0000;
    } else
    {
      rgCol.red=0xFFFF;
      rgCol.green=0x0000;
      rgCol.blue=0x0000;
    }
    gdk_gc_set_rgb_fg_color(gc, &rgCol);

    // The connection line
    gdk_gc_set_line_attributes(gc, 1, 
      GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
    gdk_draw_line(screen, gc,
      coord.getX(), coord.getY(),
      fig->getOffset().getX(), fig->getOffset().getY());

    // The enlarged ring
    gdk_gc_set_line_attributes(gc, 6,
      GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
    gdk_draw_arc(screen, gc, FALSE,
      coord.getX()-FIGURE_DOT_RADIUS, coord.getY()-FIGURE_DOT_RADIUS,
      FIGURE_DOT_RADIUS*2, FIGURE_DOT_RADIUS*2,
      0*64, 360*64);

    /* Do the dot itself */

    GdkColor dotCol;
    const guint32 rgb(getRGB(i.getColour()));
    dotCol.red=(rgb&0xFF0000)>>16;
    dotCol.red*=dotCol.red;
    dotCol.green=(rgb&0x00FF00)>>8;
    dotCol.green*=dotCol.green;
    dotCol.blue=(rgb&0x0000FF)>>0;
    dotCol.blue*=dotCol.blue;

    gdk_gc_set_line_attributes(gc, 2,
      GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
    gdk_gc_set_rgb_fg_color(gc, &dotCol);

    gdk_draw_arc(screen, gc, FALSE,
      coord.getX()-FIGURE_DOT_RADIUS, coord.getY()-FIGURE_DOT_RADIUS,
      FIGURE_DOT_RADIUS*2, FIGURE_DOT_RADIUS*2,
      0*64, 360*64);
  }
}

} // namespace gimandel
} // namespace islandmandel
