/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008,2012  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "CSet.hpp"

namespace islandmandel
{

/* ************************************************************************** */
/* ******************************** CSet ************************************ */
/* ************************************************************************** */

Figure::colourT CSet::getNumberColours() const
{
  return 2;
}

void CSet::calcArea(floatCoordT tx, floatCoordT ty,
  floatCoordT sx, floatCoordT sy, Coordinate::compT w, Coordinate::compT h,
  Figure::colourT* px, Coordinate::compT rs) const
{
  #pragma omp parallel for \
    default(none) \
    shared(tx, ty, sx, sy, w, h, px, rs)
  for(Coordinate::compT y=0; y<h; ++y)
  {
    Figure::colourT* const line(px+y*rs);
    for(Coordinate::compT x=0; x<w; ++x)
      line[x]=(diverges(ComplexNum(tx+x*sx, ty+y*sy)) ?
        COLOUR_DIV : COLOUR_CONV);
  }
}

/* ************************************************************************** */
/* ****************************** Mandelbrot ******************************** */
/* ************************************************************************** */

Mandelbrot::Mandelbrot(Coordinate::compT w, Coordinate::compT h)
  : CSet(w, h, -2.3, -1., 1., 1.)
{}

CSet::itNumT Mandelbrot::diverges(const ComplexNum& p) const
{
  /* Optimization: Check for main cardioid and first period bulb
   *
   * The main cardioid is
   *  c=(1-(m-1)**2)/4, for m*m<1
   * 4*c=1-(m-1)**2, (m-1)**2=1-4*c, m=sqrt(1-4*c)+1
   *
   * The bulb is a circle around -1 with radius 1/4.
   */
  
  // Check for main cardioid
  static const ComplexNum c1(1, 0);
  static const ComplexNum c4(4, 0);
  const ComplexNum m1(std::sqrt(c1-c4*p)+c1);
  const ComplexNum m2(-std::sqrt(c1-c4*p)+c1);
  if(std::norm(m1)<1 || std::norm(m2)<1)
    return NOT_DIVERGED;

  // Check for first period bulb
  static const ComplexNum cM1(-1, 0);
  if(std::norm(p-cM1)<=1./(4*4))
    return NOT_DIVERGED;

  // Simply iterate
  ComplexNum z(p);
  for(itNumT i(1); i!=ITERATIONS; ++i)
  {
    if(isDiverged(z))
      return i;
    z=z*z+p;
  }

  // Not diverged
  return NOT_DIVERGED;
}

/* ************************************************************************** */
/* ********************************* Julia ********************************** */
/* ************************************************************************** */

Julia::Julia(Coordinate::compT w, Coordinate::compT h,
  floatCoordT cx, floatCoordT cy)
  : CSet(w, h, -2., -2., 2., 2.), c(cx, cy)
{}

CSet::itNumT Julia::diverges(const ComplexNum& p) const
{
  // Brute force!
  ComplexNum z(p);
  for(itNumT i(1); i!=ITERATIONS; ++i)
  {
    z=z*z+c;
    if(isDiverged(z))
      return i;
  }

  // Not diverged
  return NOT_DIVERGED;
}

} // namespace islandmandel
