/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008,2012  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef IMANDEL_CSET_HPP
#define IMANDEL_CSET_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Fractal.hpp"
#include "Figure.hpp"

#include <complex>

namespace islandmandel
{

/* ************************************************************************** */
/* ******************************** CSet ************************************ */
/* ************************************************************************** */

/**
 * A fractal based on a set of complex numbers, mainly Mandelbrot and Julia.
 */
class CSet : public Fractal
{

  /* ************************************************************************ */
  public:

    /**
     * Always two colours for these fractals.
     * @see Fractal::getNumberColours()
     */
    virtual Figure::colourT getNumberColours() const;

  /* ************************************************************************ */
  protected:

    /** Type for number of iterations */
    typedef unsigned itNumT;

    /** Type used for complex numbers */
    typedef std::complex<long double> ComplexNum;

    /** Number of iterations to do */
    static const itNumT ITERATIONS=100;

    /** Iteration-number to return for non-diverging */
    static const itNumT NOT_DIVERGED=0;

    /**
     * Simply forward the constructor
     * @see Fractal::Fractal(...)
     */
    inline CSet(Coordinate::compT pw, Coordinate::compT ph,
      floatCoordT x1, floatCoordT y1, floatCoordT x2, floatCoordT y2)
      : Fractal(pw, ph, x1, y1, x2, y2)
    {}

    /**
     * Calculate the set, using diverges(...) of the sub-class.
     * @see Fractal::calcArea(...) 
     */
    virtual void calcArea(floatCoordT tx, floatCoordT ty,
      floatCoordT sx, floatCoordT sy, Coordinate::compT w, Coordinate::compT h,
      Figure::colourT* px, Coordinate::compT rs) const;

    /**
     * For the sub-classes to check whether a given number diverges (that is,
     * how many iterations it takes to diverge).
     * @param p The position to check
     * @return The number of iterations or 0 if it does not diverge
     */
    virtual itNumT diverges(const ComplexNum& p) const =0;

    /**
     * Check if a number can be regarded as diverged.
     * @param p The number to check
     * @return True iff its magnitude is big enough
     */
    static inline bool isDiverged(const ComplexNum& p)
    {
      return std::norm(p)>4; // Take 2 as abs(p)-treshold
    }

  /* ************************************************************************ */
  private:

    /** Colour code for "does not diverge" */
    static const unsigned COLOUR_CONV=0;
    /** Colour code for "does diverge" */
    static const unsigned COLOUR_DIV=1;
    // Figure::colourT gives undefined reference above

};

/* ************************************************************************** */
/* ****************************** Mandelbrot ******************************** */
/* ************************************************************************** */

/**
 * The CSet-fractal for Mandelbrot set.
 */
class Mandelbrot : public CSet
{

  /* ************************************************************************ */
  public:
    
    /**
     * Create the Mandelbrot-fractal generator for given pixel-size,
     * substituting a hardcoded minimum range.
     * @param w The pixel width
     * @param h The pixel height
     */
    Mandelbrot(Coordinate::compT w, Coordinate::compT h);

  /* ************************************************************************ */
  protected:

    /**
     * @see CSet::diverges(const ComplexNum&) const
     */
    virtual itNumT diverges(const ComplexNum& p) const;

};

/* ************************************************************************** */
/* ********************************* Julia ********************************** */
/* ************************************************************************** */

/**
 * The CSet-fractal for Julia set.
 */
class Julia : public CSet
{

  /* ************************************************************************ */
  public:
    
    /**
     * Create the Julia-fractal generator for given pixel-size,
     * substituting a hardcoded minimum range.
     * @param w The pixel width
     * @param h The pixel height
     * @param cx The parameter real part
     * @param cy The parameter imaginary part
     */
    Julia(Coordinate::compT w, Coordinate::compT h,
      floatCoordT cx, floatCoordT cy);

  /* ************************************************************************ */
  protected:

    /**
     * @see CSet::diverges(const ComplexNum&) const
     */
    virtual itNumT diverges(const ComplexNum& p) const;

  /* ************************************************************************ */
  private:

    /** The parameter */
    const ComplexNum c;

};

} // namespace islandmandel

#endif // Header-guard
