/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Fractal.hpp"

#include <cassert>
#include <algorithm>
#include <iostream>

// Output modifications done with fractal?
//#define LOG_MODIFICATIONS

namespace islandmandel
{

/* ************************************************************************** */
/* **************************** Fractal ************************************* */
/* ************************************************************************** */

const Figure::colourT Fractal::DUMMY_COLOUR(-1);

Fractal::Fractal(Coordinate::compT pw, Coordinate::compT ph,
  floatCoordT x1, floatCoordT y1, floatCoordT x2, floatCoordT y2)
  : pixelWidth(pw), pixelHeight(ph), cx((x1+x2)/2), cy((y1+y2)/2),
    pixels(NULL)
{
  // Calculate ideal step in both directions first
  assert(x2>x1);
  stepX=(x2-x1)/pixelWidth;
  assert(stepX>0);
  assert(y2>y1);
  stepY=(y2-y1)/pixelHeight;
  assert(stepY>0);

  // Now, choose larger of those two generally
  if(stepY>stepX)
    stepX=stepY;
  else
    stepY=stepX;
  assert(stepX==stepY);
}

Fractal::~Fractal()
{
  if(pixels)
    delete pixels;
}

Figure::colourT Fractal::getPixelColour(const Coordinate& c) const
{
  // Out of range?
  if(c.getX()<0 || c.getX()>=getWidth())
    return DUMMY_COLOUR;
  if(c.getY()<0 || c.getY()>=getHeight())
    return DUMMY_COLOUR;

  return getPixels()[c.getY()*getRowstride()+c.getX()];
}

void Fractal::moveZoom(Coordinate::compT dx, Coordinate::compT dy,
  floatCoordT f)
{
#ifdef LOG_MODIFICATIONS
  std::cout << "movezoom" << " " << dx << " " << dy << " " << f << std::endl;
#endif

  cx+=dx*stepX;
  cy+=dy*stepY;
  stepX/=f;
  stepY/=f;

  if(pixels)
  {
    delete pixels;
    pixels=NULL;
  }

  calculate();
}

void Fractal::zoom(floatCoordT f)
{
#ifdef LOG_MODIFICATIONS
  std::cout << "zoom" << " " << f << std::endl;
#endif

  stepX/=f;
  stepY/=f;

  if(pixels)
  {
    delete pixels;
    pixels=NULL;
  }

  calculate();
}

void Fractal::move(Coordinate::compT dx, Coordinate::compT dy)
{
#ifdef LOG_MODIFICATIONS
  std::cout << "move" << " " << dx << " " << dy << std::endl;
#endif

  cx+=dx*stepX;
  cy+=dy*stepY;
  calculate();
}

const Figure::colourT* Fractal::getPixels() const
{
  assert(pixels);
  return pixels->getPixels(getRealX(0), getRealY(0), pixelWidth, pixelHeight);
}

Coordinate::compT Fractal::getRowstride() const
{
  assert(pixels);
  return pixels->getRowstride();
}

void Fractal::calculate()
{
  const floatCoordT tx(getRealX(0));
  const floatCoordT ty(getRealY(0));

  if(!pixels)
    pixels=new PixelCache(*this, tx, ty, stepX, stepY, pixelWidth, pixelHeight);
  else
    pixels->ensureArea(tx, ty, pixelWidth, pixelHeight);
}

/* ************************************************************************** */
/* **************************** Fractal::PixelCache ************************* */
/* ************************************************************************** */

Fractal::PixelCache::PixelCache(const Fractal& f,
  floatCoordT tx, floatCoordT ty, floatCoordT sx, floatCoordT sy,
  Coordinate::compT pw, Coordinate::compT ph)
  : fract(f),
    topX(tx), topY(ty), stepX(sx), stepY(sy),
    pixelWidth(pw), pixelHeight(ph),
    pixels(new Figure::colourT[pixelWidth*pixelHeight])
{
  assert(pixels);
  std::fill(pixels, pixels+pixelWidth*pixelHeight, Fractal::DUMMY_COLOUR);
  ensureArea(topX, topY, pixelWidth, pixelHeight);

  assert(stepX==fract.stepX && stepY==fract.stepY);
}

Fractal::PixelCache::~PixelCache()
{
  assert(pixels);
  delete[] pixels;
}

const Figure::colourT* Fractal::PixelCache::getPixels(
  floatCoordT tx, floatCoordT ty,
  Coordinate::compT pw, Coordinate::compT ph) const
{
  assert(stepX==fract.stepX && stepY==fract.stepY);

  assert(pw<=pixelWidth && ph<=pixelHeight);
  assert(topX<=tx && topY<=ty);

  // Calculate the offsets into our pixel map
  const Coordinate::compT pixelX((tx-topX)/stepX);
  const Coordinate::compT pixelY((ty-topY)/stepY);

  return pixels+getRowstride()*pixelY+pixelX;
}

void Fractal::PixelCache::ensureArea(floatCoordT tx, floatCoordT ty,
  Coordinate::compT pw, Coordinate::compT ph)
{
  assert(stepX==fract.stepX && stepY==fract.stepY);

  // Ensure left/top edges are far enough away
  while(topX>tx)
    resizeX(false);
  while(topY>ty)
    resizeY(false);

  // Calculate the offsets into our pixel map
  assert(topX<=tx);
  assert(topY<=ty);
  const Coordinate::compT pixelX((tx-topX)/stepX);
  const Coordinate::compT pixelY((ty-topY)/stepY);

  // Ensure right/bottom edges are far enough away
  while(pixelWidth<pw+pixelX)
    resizeX(true);
  while(pixelHeight<ph+pixelY)
    resizeY(true);
  assert(pw+pixelX<=pixelWidth);
  assert(ph+pixelY<=pixelHeight);

  // Peek through the data, make sure each pixel is calculated
  for(Coordinate::compT y(pixelY); y!=pixelY+ph; ++y)
    for(Coordinate::compT x(pixelX); x!=pixelX+pw; ++x)
    {
      if(pixels[y*pixelWidth+x]==Fractal::DUMMY_COLOUR)
        calcRect(x, y, pixelX+pw-x, pixelY+ph-y);
      assert(pixels[y*pixelWidth+x]!=Fractal::DUMMY_COLOUR);
    }
}

void Fractal::PixelCache::calcRect(Coordinate::compT tx, Coordinate::compT ty,
  Coordinate::compT pw, Coordinate::compT ph)
{
  assert(stepX==fract.stepX && stepY==fract.stepY);
  assert(tx+pw<=pixelWidth);
  assert(ty+ph<=pixelHeight);

  // Find size of needed rectangle
  Coordinate::compT neededWidth, neededHeight;
  for(neededWidth=1; neededWidth!=pw; ++neededWidth)
    if(pixels[pixelWidth*ty+tx+neededWidth]!=Fractal::DUMMY_COLOUR)
      break;
  for(neededHeight=1; neededHeight!=ph; ++neededHeight)
  {
    bool nextLineOk(true);
    for(Coordinate::compT x(tx); x!=tx+neededWidth; ++x)
      if(pixels[pixelWidth*(ty+neededHeight)+x]!=Fractal::DUMMY_COLOUR)
      {
        nextLineOk=false;
        break;
      }
    if(!nextLineOk)
      break;
  }

  // Calculate tx/ty in real coordinates
  const floatCoordT txReal(topX+stepX*tx);
  const floatCoordT tyReal(topY+stepY*ty);

  // Call the fractal to calculate
  fract.calcArea(txReal, tyReal, stepX, stepY, neededWidth, neededHeight,
    pixels+pixelWidth*ty+tx, pixelWidth);
}

void Fractal::PixelCache::resizeY(bool bottom)
{
  // Allocate and initialize the new memory
  Figure::colourT* pxNew(new Figure::colourT[pixelWidth*pixelHeight*2]);
  assert(pxNew);
  std::fill(pxNew, pxNew+pixelWidth*pixelHeight*2, Fractal::DUMMY_COLOUR);

  // Copy the existing data over
  std::copy(pixels, pixels+pixelWidth*pixelHeight,
    pxNew+(bottom ? 0 : pixelWidth*pixelHeight));

  // If resized at the top, update top y
  if(!bottom)
    topY-=pixelHeight*stepY;

  // Take the new data over
  delete[] pixels;
  pixels=pxNew;
  pixelHeight*=2;
}

void Fractal::PixelCache::resizeX(bool right)
{
  // Allocate and initialize the new memory
  Figure::colourT* pxNew(new Figure::colourT[pixelWidth*pixelHeight*2]);
  assert(pxNew);
  std::fill(pxNew, pxNew+pixelWidth*pixelHeight*2, Fractal::DUMMY_COLOUR);

  // Copy the existing data over line-by-line
  const Coordinate::compT oldRS(pixelWidth);
  const Coordinate::compT newRS(pixelWidth*2);
  for(Coordinate::compT row(0); row!=pixelHeight; ++row)
    std::copy(pixels+oldRS*row, pixels+oldRS*(row+1),
      pxNew+newRS*row+(right ? 0 : pixelWidth));

  // If resized at the left, update top x
  if(!right)
    topX-=pixelWidth*stepX;

  // Take the new data over
  delete[] pixels;
  pixels=pxNew;
  pixelWidth*=2;
}

} // namespace islandmandel
