/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef IMANDEL_FRACTAL_HPP
#define IMANDEL_FRACTAL_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Figure.hpp"
#include "Geometry.hpp"

namespace islandmandel
{

/* ************************************************************************** */
/* **************************** Fractal ************************************* */
/* ************************************************************************** */

/**
 * The fractal as game-map itself.  This is the generic interface handling
 * zooming/scaling/caching and calls further into the specific implementation
 * to fill in a rectangular area with pixel data.
 * Because of the zooming it works with floating-point coordinates to specify
 * sub-rectangles when interfacing the fractal generator, but has also a
 * coordinate-system in pixels for interfacing the other end.
 */
class Fractal
{

  /* ************************************************************************ */
  public:

    /** Internal floating point coordinate type */
    typedef long double floatCoordT;

    /** Colour-code used to mark a dummy point in the pixel data */
    static const Figure::colourT DUMMY_COLOUR;

    /**
     * Destroy, possibly for subclasses, too.
     */
    virtual ~Fractal();

    // Query for image data, rowstride and width/height
    const Figure::colourT* getPixels() const;
    Coordinate::compT getRowstride() const;
    inline Coordinate::compT getWidth() const
    {
      return pixelWidth;
    }
    inline Coordinate::compT getHeight() const
    {
      return pixelHeight;
    }

    // Map pixel-coordinates to real coordinates
    inline floatCoordT getRealX(Coordinate::compT px) const
    {
      assert(px>=0);
      assert(px<getWidth());
      return cx+stepX*(px-pixelWidth/2);
    }
    inline floatCoordT getRealY(Coordinate::compT py) const
    {
      assert(py>=0);
      assert(py<getHeight());
      return cy+stepY*(py-pixelHeight/2);
    }

    /**
     * Query for the colour of a certain pixel in pixel-coordinates, for
     * checking the figure placement.
     * @param c The coordinate in pixels
     * @return The colour of this pixel
     */
    Figure::colourT getPixelColour(const Coordinate& c) const;

    /**
     * Query for the number of colours in the fractal type
     * @return The number of colours used in this fractal
     */
    virtual Figure::colourT getNumberColours() const =0;

    /**
     * Zoom in/out with given factor
     * @param f The zoom-factor
     */
    void zoom(floatCoordT f);

    /**
     * Move the area by pixel coordinates
     * @param dx Distance in x in pixels
     * @param dy Distance in y in pixels
     */
    void move(Coordinate::compT dx, Coordinate::compT dy);

    /**
     * Do a move followed by zooming directly
     * @param dx Distance to move x
     * @param dy Distance to move y
     * @param f The zoom factor
     */
    void moveZoom(Coordinate::compT dx, Coordinate::compT dy, floatCoordT f);

    /**
     * Ensure the currently chosen area is present in the pixel data.
     */
    void calculate();

  /* ************************************************************************ */
  protected:

    /**
     * Create for sub-classes, with requested minimum starting range and
     * adapted for the correct needed pixel sizes.
     * @param pw Pixel-width needed
     * @param ph Pixel-height needed
     * @param x1 Requested left edge in floating-point coordinates
     * @param y1 Requested top edge
     * @param x2 Requested right edge
     * @param y2 Requested bottom edge
     */
    Fractal(Coordinate::compT pw, Coordinate::compT ph,
      floatCoordT x1, floatCoordT y1, floatCoordT x2, floatCoordT y2);

    /**
     * Call subclass to calculate a given area and fill in the pixel data
     * @param tx Top-x in floating-point
     * @param ty Top-y in floating-point
     * @param sx Step-x in floating-point
     * @param sy Step-y in floating-point
     * @param w Width in pixels
     * @param h Height in piels
     * @param px Fill in here
     * @param rs Rowstride of the pixel-array
     */
    virtual void calcArea(floatCoordT tx, floatCoordT ty,
      floatCoordT sx, floatCoordT sy, Coordinate::compT w, Coordinate::compT h,
      Figure::colourT* px, Coordinate::compT rs) const =0;

  /* ************************************************************************ */
  private:

    class PixelCache;

    // Width and height of area in pixels
    Coordinate::compT pixelWidth;
    Coordinate::compT pixelHeight;

    // Center of the current area in floating-point coordinates
    floatCoordT cx;
    floatCoordT cy;

    // Pixel-sizes of the current area in floating-points
    floatCoordT stepX;
    floatCoordT stepY;

    /** Current pixel data cache */
    PixelCache* pixels;

};

/* ************************************************************************** */
/* **************************** Fractal::PixelCache ************************* */
/* ************************************************************************** */

/**
 * Stores the already calculated pixel data of a Fractal similar to a
 * two-dimensional vector-data-structure, expanding the rectangle when needed.
 * It does recalculation by means of the associated Fractal when some parts of
 * a needed area are missing.
 */
class Fractal::PixelCache
{

  /* ************************************************************************ */
  public:

    /**
     * Construct a new pixel cache, it is initialized for the given area.
     * @param f The fractal
     * @param tx Top-x
     * @param ty Top-y
     * @param sx Step-x
     * @param sy Step-y
     * @param pw Pixel width
     * @param ph Pixel height
     */
    PixelCache(const Fractal& f, floatCoordT tx, floatCoordT ty,
      floatCoordT sx, floatCoordT sy,
      Coordinate::compT pw, Coordinate::compT ph);

    /**
     * Destroy, freeing the pixeldata
     */
    ~PixelCache();

    /**
     * Query for the pixel data pointer for a rectangle
     * @param tx Top-x of the rectangle in question
     * @param ty Top-y
     * @param pw Pixel width of it
     * @param ph Pixel height
     */
    const Figure::colourT* getPixels(floatCoordT tx, floatCoordT ty,
      Coordinate::compT pw, Coordinate::compT ph) const;

    /**
     * Query for the rowstride for our pixel datas
     */
    inline Coordinate::compT getRowstride() const
    {
      return pixelWidth;
    }

    /**
     * Ensure a particular area is calculated, calculate if not.
     * @param tx Top-x of the rectangle wanted
     * @param ty Top-y
     * @param pw Pixel width
     * @param ph Pixel height
     */
    void ensureArea(floatCoordT tx, floatCoordT ty,
      Coordinate::compT pw, Coordinate::compT ph);

  /* ************************************************************************ */
  private:

    /** The associated fractal */
    const Fractal& fract;

    // The top-left corner of our rectangle in real coordinates
    floatCoordT topX;
    floatCoordT topY;

    // Our step-distance in real coordinates per pixel
    const floatCoordT stepX;
    const floatCoordT stepY;

    // Width/height of the allocated rectangle in pixels
    Coordinate::compT pixelWidth;
    Coordinate::compT pixelHeight;

    /** Allocated pixel data row-by-row */
    Figure::colourT* pixels;

    // Reallocate doubling into the given direction
    void resizeX(bool right);
    void resizeY(bool bottom);

    /**
     * Find and calculate the uncalculated rectangle at the given point.
     * @param tx Start here x
     * @param ty Start here y
     * @param pw Maximum width to calculate
     * @param ph Maximum height to calculate
     */
    void calcRect(Coordinate::compT tx, Coordinate::compT ty,
      Coordinate::compT pw, Coordinate::compT ph);

};

} // namespace islandmandel

#endif // Header-guard
