/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008,2012  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

/* Basic test for CSet-fractals (Mandelbrot), so far possible.  Calculates one
 * with low resulution, prints it in "Ascii-art", and checks number of black
 * points against hard-coded "correct" value.
 * Also checks the scaling/moving features of Fractal, scaling in some way until
 * all should be black and then moving so it should all be white.
 */

#include "Test.hpp"

#include "Figure.hpp"
#include "CSet.hpp"

#include <cstdlib>
#include <iostream>

using namespace islandmandel;
using namespace tests;

// Size of the output to generate
#define WIDTH 80
#define HEIGHT 20

// Colours for black and white, respectively
#define COLOUR_BLACK 0
#define COLOUR_WHITE 1

/**
 * Count the black dots and print it optionally on the way
 * @param f The fractal
 * @param print Whether to print it as Ascii-art on the way
 * @return The number of black points
 */
unsigned countBlack(const Fractal& f, bool print =false)
{
  unsigned result(0);
  for(Coordinate::compT y(0); y!=HEIGHT; ++y)
  {
    const Figure::colourT* const line(f.getPixels()+y*f.getRowstride());
    for(Coordinate::compT x(0); x!=WIDTH; ++x)
    {
      check(line[x]==COLOUR_BLACK || line[x]==COLOUR_WHITE,
        "Completely wrong colour code in Mandelbrot-set!");
      if(line[x]==COLOUR_BLACK)
      {
        ++result;
        if(print)
          std::cout << "*";
      } else
        if(print)
          std::cout << " ";
    }
    if(print)
      std::cout << std::endl;
  }
  assert(result<=WIDTH*HEIGHT);
  return result;
}

int main()
{
  // Create the fractal
  Mandelbrot fract(WIDTH, HEIGHT);
  fract.calculate();

  // Count the black dots, printing it, and compare to magic number
  unsigned blacks(countBlack(fract, false));
  check(blacks==157,
   "There should be 157 black dots (not %u), shouldn't there?", blacks);

  // Zoom in dramatically, all should be black
  fract.zoom(100);
  blacks=countBlack(fract);
  check(blacks==WIDTH*HEIGHT, "All should be black!");

  // Move a bit, all should become white
  fract.move(100*WIDTH, 0);
  blacks=countBlack(fract);
  check(blacks==0, "All should be white!");

  return EXIT_SUCCESS;
}
